from abc import ABC, abstractmethod
import pickle
import hashlib
import logging
import esipysi

logger = logging.getLogger("EsiPysi")

class EsiCache(ABC):
    """
    This is an abstract class used as a template for building caching systems
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def store(self, operation_id, operation_parameters, value, cache_time = -1):
        """
        Stores data in the cache

        :param operation_id: The id of the operation
        :param operation_parameters: parameters for the operation
        :type operation_parameters: dict
        :param value: The value to be stored in the cache
        :param cache_time: How long to store in cache (-1 is unlimited)
        :type cache_time: int (seconds)
        """
        raise NotImplementedError()

    @abstractmethod
    def in_cache(self, operation_id, operation_parameters):
        """
        Check for a cache hit

        :param operation_id: The id of the operation
        :param operation_parameters: parameters for the operation
        :type operation_parameters: dict

        :return: True if stored in the cache, False otherwise
        """
        raise NotImplementedError()

    @abstractmethod
    def retrieve(self, operation_id, operation_parameters, default=None):
        """
        Returns data in the cache

        :param operation_id: The id of the operation
        :param operation_parameters: parameters for the operation
        :type operation_parameters: dict
        :param default: A default value if not in cache (defaults to None)

        :return: stored value if in the cache, default otherwise
        """
        raise NotImplementedError()

    def get_key(self, operation_id, operation_parameters):
        """
        Returns the key used in the cache

        :param operation_id: The id of the operation
        :param operation_parameters: parameters for the operation
        :type operation_parameters: dict

        :return: The key generated to store the value in the cache
        """
        #Generate a key
        hash_dict = {}
        hash_dict["cache_operation_id"] = operation_id
        hash_dict["esipysi_version"] = esipysi.__version__
        for key, value in operation_parameters.items():
            hash_dict[key] = pickle.dumps(value)

        key_set = frozenset(hash_dict.items())
        hash = hashlib.md5()
        key_pickle = pickle.dumps(key_set)
        hash.update(key_pickle)
        key = hash.hexdigest()
        logger.debug("hashed {} to {}".format(key_pickle, key))
        return key


import asyncio
import aiohttp
import requests
import json
from .op import EsiOp
from .auth import EsiAuth
from .cache.cache import EsiCache
import logging

logger = logging.getLogger("EsiPysi")

class EsiPysi(object):
    """
    The EsiPysi class creates "EsiOp" operations based on a provided swagger spec
    """

    def __init__(self, swagger_url, **kwargs):
        """
        Initialize the class

        Arguments:
            swagger_url -- Url to the swagger spec

        Keyword arguments:
            user_agent -- user agent to send with ESI calls
            cache -- EsiCache object to use for caching
            auth -- EsiAuth to use for authorized calls to ESI
            loop -- Event loop to use for asyncio
        """
        self.args = kwargs
        self.__is_ready = False
        self.__loading = False

        cache = kwargs.get("cache")
        if cache is not None:
            if not issubclass(EsiCache, type(cache)):
                TypeError("cache should be of the type EsiCache")

        self.operations = {}
        self.data = {}
        
        r = requests.get(swagger_url)
        data = r.json()
            
        self.data = data
        self.__analyze_swagger()
        self.__is_ready = True


    def __analyze_swagger(self):
        #Get base url
        self.base_url = "https://" + self.data.get("host","") + self.data.get("basePath", "")

        #Reformat json
        paths = self.data.get("paths", {})
        #each path
        for route, verbs in paths.items():
            #each http verb in a path
            for verb, operation in verbs.items():
                operation_id = operation.get("operationId")
                if operation_id is None:
                    continue
                new_op = operation.copy()
                new_op["path"] = route
                new_op["verb"] = verb

                #Handle parameter refs
                params = operation.get("parameters")
                new_op["parameters"] = {}
                for param in params:
                    path = param.get("$ref")
                    if path is None:
                        param_details = param.copy()
                    else:
                        param_details = self.__get_ref(path)
                    param_name = param_details.get("name")
                    new_op["parameters"][param_name] = param_details

                self.operations[operation_id] = new_op

    def __get_ref(self, path):
        path_split = path.split("/")
        if path_split[0] != "#":
            #Unsupported
            return None
        ref = self.data
        for i in range(1, len(path_split)):
            ref = ref.get(path_split[i], {})

        return ref

    def get_operation(self, operation_id):
        """
        Get an ESI operation from it's id

        Arguments:
            operation_id -- The id of the swagger operation
        """
        operation = self.operations.get(operation_id)
        if operation is None:
            return None
        return EsiOp(operation, self.base_url, **self.args)


import asyncio
import aiohttp

import re
import urllib
import json
from urllib.error import HTTPError
from json.decoder import JSONDecodeError
from .cache.cache import EsiCache
from .auth import EsiAuth
from urllib.parse import urlencode
import logging

logger = logging.getLogger("EsiPysi")

class EsiOp(object):
    """ A class to handle operations of the ESI api """
    def __init__(self, operation, base_url, **kwargs):
        """
        Initialize the class - this should only be done through an EsiPysi class object

        Arguments:
            session -- A ClientSession from aiohttp used to make ESI calls
            operation -- A dict with the modified swagger structure that defines the operation
            base_url -- url to the base endpoint
        Keyword arguments:
            user_agent -- User agent to use when interacting with ESI
            cache -- an EsiCache object to use for caching queries
            auth -- an EsiAuth object to use for authorizing the ESI call
            loop -- A asyncio event loop to use for async calls
        """
        self.__operation = operation
        self.__path = operation.get("path")
        self.__verb = operation.get("verb")
        self.__parameters = operation.get("parameters")
        self.__operation_id = operation.get("operationId")
        self.__base_url = base_url
        self.__cached_seconds = operation.get("x-cached-seconds", -1)
        self.user_agent = kwargs.get("user_agent")
        self.cache = kwargs.get("cache")
        auth_args = kwargs.get("auth")
        self.__auth = None
        self.set_auth(auth_args)

        self.__loop = kwargs.get("loop")
        self.__use_loop = True
        if self.__loop is None:
            self.__use_loop = False
        elif not isinstance(self.__loop, asyncio.BaseEventLoop):
            raise TypeError("loop must be a asyncio event loop")

        self.use_cache = False
        if self.cache is not None:
            if not issubclass(EsiCache, type(self.cache)):
                ValueError("cache should be of the type EsiCache")
            self.use_cache = True

    def set_auth(self, esiauth):
        """
        Set the authorization for this operation

        Arguments:
            esiauth -- An EsiAuth object which contains the authorization info
        """
        if not issubclass(EsiAuth, type(esiauth)):
            ValueError("esiauth should be of the type EsiAuth")
        
        self.auth = esiauth

    async def json(self, **kwargs):
        """
        Call the ESI API and retrieve the json data and decode as a dict

        Keyword arguments:
            Arguments from the ESI call
        """
        return await self.__call_esi_async(False, **kwargs)

    async def raw(self, **kwargs):
        """
        Call the ESI API and retrieve the raw result

        Keyword arguments:
            Arguments from the ESI call
        """
        return await self.__call_esi_async(True, **kwargs)

    async def __call_esi_async(self, raw, **kwargs):
        if self.__loop is None:
            async with aiohttp.ClientSession() as session:
                return await self.__execute(session, raw, **kwargs)
        else:
            async with aiohttp.ClientSession(loop = self.__loop) as session:
                return await self.__execute(session, raw, **kwargs)
        

    async def __execute(self, session, raw, **kwargs):

        cache_value = self.__cache_read(raw, **kwargs)
        if cache_value is not None:
            return cache_value

        url = self.__base_url + self.__path
        #Handle parameters
        for name, param in self.__parameters.items():
            required = param.get("required", False)
            if required:
                if name not in kwargs:
                    raise ValueError("'{}' is a required parameter".format(name))

        body = None
        query_parameters = {}
        headers = {}

        for key, value in kwargs.items():
            if key not in self.__parameters:
                raise ValueError("'{}' is not a valid parameter".format(key))
            param_info = self.__parameters.get(key)
            #Handle path parameters
            param_type = param_info.get("in")
            if param_type == "path":
                escaped_value = urllib.parse.quote(str(value))
                url = url.replace("{" + param_info.get("name") + "}", escaped_value)
            elif param_type == "query":
                query_parameters[key] = value
            elif param_type == "header":
                headers[key] = value
            elif param_type == "body":
                body = json.dumps(value)

        if self.user_agent is not None:
            headers['User-Agent'] = self.user_agent
        if self.auth is not None:
            auth_code = await self.auth.authorize()
            headers["Authorization"] = "Bearer {}".format(auth_code)
        #Call operation
        logger.info("Calling '{}' with data '{}' using HTTP {}".format(url, body, self.__verb.upper()))

        if self.__verb.lower() == "get":
            async with session.get(url, params=query_parameters, headers=headers) as resp:
                return await self.__process_response(resp, raw, **kwargs)
        else:
            #After this I don't know WTF CCP was thinking
            if query_parameters:
                query_str = "?" + urlencode(query_parameters)
            else:
                query_str = ""
            full_url = url + query_str
            if self.__verb.lower() == "post":
                async with session.post(full_url, data=body, headers=headers) as resp:
                    return await self.__process_response(resp, raw, **kwargs)
            elif self.__verb.lower() == "put":
                async with session.put(full_url, data=body, headers=headers) as resp:
                    return await self.__process_response(resp, raw, **kwargs)
            elif self.__verb.lower() == "delete":
                async with session.delete(full_url, data=body, headers=headers) as resp:
                    return await self.__process_response(resp, raw, **kwargs)
        
    async def __process_response(self, resp, raw, **kwargs):
        text = await resp.text()
        if resp.status >= 400:
            exception = HTTPError(resp.url, resp.status, text, resp.headers, None)
            logger.exception("ESI HTTP error occured: {}".format(exception))
            raise exception

        if self.use_cache:
            self.cache.store(self.__operation_id, kwargs, text, self.__cached_seconds)

        if raw:
            return text

        try:
            return await resp.json()
        except Exception:
            logger.exception("Could not parse JSON")
            return None

    def __cache_read(self, raw, **kwargs):
        if not self.use_cache:
            return None
        
        value = self.cache.retrieve(self.__operation_id, kwargs)
        result = None
        if value is not None:
            if raw:
                result = value
            else:
                try:
                    result = json.loads(value)
                except Exception:
                    logger.exception("Could not parse JSON")
            if result is not None:
                logger.info("Result found in cache for {} : {}".format(self.__operation_id, kwargs))
        return result


    def __str__(self):
        return self.__operation_id
        


import datetime
import asyncio
import aiohttp
import requests
import base64
from urllib.error import HTTPError
import logging

logger = logging.getLogger("EsiPysi")

class EsiAuth(object):

    """
    Keeps track of authorization information, pass to an EsiOp to authorize
    """
    def __init__(self, client_id, client_secret, access_token, refresh_token, expires_at, login_server="login.eveonline.com", loop=None):
        """
        Sets up the authorization

        Arguments:
            client_id -- Eve 3rd party developer client id
            client_secret -- The secret key to go with the client_id
            access_token -- Access token from Eve SSO
            refresh_token -- Refresh token from Eve SSO
            expires_at -- a UTC datetime of when the access token will expire
            login_server -- The Eve SSO login server (defaults to tranquility login server)
            loop -- An asyncio loop for esi calls (defaults to the current loop)
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires_at = expires_at
        if loop is None:
            self.loop = None
        else:
            if isinstance(loop, asyncio.BaseEventLoop):
               self.loop = loop
            else:
                raise TypeError("Loop argument must either be 'None' or impliment BaseEventLoop") 

        self.__login_server = login_server

        key_secret = '{}:{}'.format(client_id, client_secret).encode('ascii')
        b64_encoded_key = base64.b64encode(key_secret)
        self.basic_auth = b64_encoded_key.decode('ascii')

    @classmethod
    async def from_authorization_code(cls, client_id, client_secret, authorization_code, login_server="login.eveonline.com"):
        esi_auth = cls(client_id, client_secret, None, None, None, login_server)
        await esi_auth.get_new_token(authorization_code)
        return esi_auth

    @classmethod 
    async def from_refresh_token(cls, client_id, client_secret, refresh_token, login_server="login.eveonline.com"):
        esi_auth = cls(client_id, client_secret, None, refresh_token, None, login_server)
        await esi_auth.get_new_token()
        return esi_auth

    async def authorize(self):
        """
        Returns the access token to be used in authorizations.  This function is reccomended over 
        getting the access token from the properties because it checks if the token is expired.

        Returns:
            Access token to ESI
        """
        if(self.is_expired()):
            await self.get_new_token()
        return self.access_token

    def is_expired(self, offset=0):
        """
        Determines if the access token is expired

        Arguments:
            offset -- seconds to shift the check, e.g. with an offset of 30 if the token is still valid for 29 seconds the check will fail (default: 0)
        
        Returns:
            True if token is valid, False otherwise
        """
        if self.expires_at is None:
            return True
        offset_delta = datetime.timedelta(seconds=offset)
        return (self.expires_at + offset_delta) <= datetime.datetime.utcnow()

    async def get_new_token(self, authorization_code = None):
        """
        Acquire a new token using the refresh token or provided authorization_code

        Arguments:
            authorization_code -- The code from the Authorization Code Grant of Eve SSO, only required if a refresh token is not avaliable (default: None)
        """    
        if self.refresh_token is not None:
            data = {"grant_type":"refresh_token","refresh_token": self.refresh_token}
        elif authorization_code is not None:
            data = {"grant_type":"authorization_code","code": authorization_code}
        else:
            raise ValueError("Either refresh_token or authorization_code is required")

        url = "https://" + self.__login_server + "/oauth/token"
        headers = {"Authorization" : "Basic {}".format(self.basic_auth), "Content-Type": "application/x-www-form-urlencoded"}


        async with aiohttp.ClientSession(loop=self.loop) as session:
            async with session.post(url, headers=headers, data=data) as resp:
                text = await resp.text()
                if resp.status >= 400:
                    raise HTTPError(url, resp.status, text, headers, None)

                result = await resp.json()
                self.access_token = result.get("access_token")
                self.refresh_token = result.get("refresh_token") #*Should* be the same but let's grab it anyway
                expires_in = int(result.get("expires_in", "0"))
                self.expires_at = datetime.datetime.utcnow() + datetime.timedelta(seconds = expires_in)

    async def verify(self, raw = False):
        """
        Retrieve details about the authorized user / token
        
        Arguments:
            raw -- Output directly as text instead of parsing the JSON

        Returns:
            Results of the verify SSO call
        """

        url = "https://" + self.__login_server + "/oauth/verify"
        auth = await self.authorize()
        headers = {"Authorization" : "Bearer {}".format(auth)}
        async with aiohttp.ClientSession(loop=self.loop) as session:
            async with session.get(url, headers=headers) as resp:
                text = await resp.text()
                if resp.status >= 400:
                    raise HTTPError(url, resp.status, text, headers, None)

                if raw:
                    return text
                return await resp.json()
        


        



        


from .cache import EsiCache
import pickle
try:
    from redis import StrictRedis
except ImportError:
    #Redis not installed
    pass


class RedisCache(EsiCache):

    def __init__(self, redis_client):
        """
        :param redis_client: The redis client to store the cache in
        :type redis_client: StrictRedis
        """
        if not issubclass(StrictRedis, type(redis_client)):
            raise ValueError("redis_client should be a StrictRedis object")

        self.redis = redis_client

    def store(self, operation_id, operation_parameters, value, cache_time = -1):
        key = self.get_key(operation_id, operation_parameters)

        self.redis.set(key, pickle.dumps(value))
        if cache_time != -1:
            self.redis.expire(key, cache_time)

    def in_cache(self, operation_id, operation_parameters):
        key = self.get_key(operation_id, operation_parameters)

        return self.redis.exists(key)

    def retrieve(self, operation_id, operation_parameters, default=None):
        key = self.get_key(operation_id, operation_parameters)

        value = self.redis.get(key)
        if value is None:
            return default
        return pickle.loads(value)




from .cache import EsiCache

class MockCache(EsiCache):

    def __init__(self):
        """
            Chache used for testing, do not use for production as it lacks the ability to expire keys
        """
        self.cache = {}

    def store(self, operation_id, operation_parameters, value, cache_time = -1):
        key = self.get_key(operation_id, operation_parameters)

        self.cache[key] = value

    def in_cache(self, operation_id, operation_parameters):
        key = self.get_key(operation_id, operation_parameters)
        return key in self.cache

    def retrieve(self, operation_id, operation_parameters, default=None):
        key = self.get_key(operation_id, operation_parameters)

        return self.cache.get(key)